/**
 *  Plugin.cpp
 *
 *  MailerQ plugin for SMTP authentication
 *
 *  @copyright 2016 Copernica B.V.
 */

// @todo remove debug code
#include <iostream>

/**
 *  Dependencies
 */
#include <libfastjson/json.h>
#include <mailerq.h>
#include "user.h"

/**
 *  These functions will be available
 *  in plain C, so export them as such
 */
extern "C" {

/**
 *  The MailerQ plugin version we support
 *
 *  @return integer
 */
MQ_EXPORT int mq_version()
{
    // return the version we compiled with
    return MQ_API_VERSION;
}

/**
 *  This function is called by MailerQ for every worker thread that is created
 *  @param  context
 */
MQ_EXPORT bool mq_context_initialize(MQ_Context *context) 
{
    // return false, MailerQ can stay in control
    return false;
}

/**
 *  This function is called by MailerQ for every worker thread
 *  that comes to an end
 */
MQ_EXPORT bool mq_context_cleanup(MQ_Context *ctxt)
{
    // return false, MailerQ can stay in control
    return false;
}

/**
 *  Initialize an incoming connection
 *
 *  @param  context     The context processing the connection
 *  @param  connection  The fresh connection
 *  @return do we take responsibility here
 */
MQ_EXPORT bool mq_smtp_in_connect(MQ_Context *context, MQ_Connection *connection)
{
    // we don't have any data yet that has to be freed
    MQ_setConnectionData(connection, nullptr);

    // we do not take responsibility here
    return false;
}

/**
 *  We would like to authenticate SMTP clients
 *
 *  @param  context     The context processing the connection
 *  @param  connection  The connection requesting authentication
 *  @param  login       The username the client tries to login with
 *  @param  loginsize   The number of characters in the username
 *  @param  passwd      The password the client tries to login with
 *  @param  passwdsize  The number of characters in the password
 *  @return do we take responsibility here
 */
MQ_EXPORT bool mq_smtp_in_authenticate(MQ_Context *context, MQ_Connection *connection, const char *login, size_t loginsize, const char *passwd, size_t passwdsize)
{
    // prevent exceptions
    try
    {
        // get user from the password database
        auto *user = new User(login, loginsize);
        
        // store the user-data with this connection
        MQ_setConnectionData(connection, user);

        // authenticate the user
        if (!user->matches(std::string(passwd, passwdsize))) throw std::runtime_error("Invalid password");

        // authenticate the user
        MQ_setAuthenticated(connection, true);
    }
    catch (const std::runtime_error &error)
    {
        // reconstruct the full error message
        std::string message("535 ");
        
        // append the rest
        message.append(error.what()).append("\r\n");
        
        // send back an error message
        MQ_send(connection, message.data(), message.size());
    }
        
    // this call does not do any operations that are in progress, so we return false to tell mailerq that it is back in control
    return false;
}

/**
 *  Process an incoming RCPT TO: instruction
 *  
 *  @param  context
 *  @param  connection
 *  @param  recipient
 *  @return bool
 */
bool mq_smtp_in_rcptto(MQ_Context *context, MQ_Connection *connection, MQ_Address *recipient)
{
        
    // if mailerq already considers this recipient to be local the message will be accepted 
    if (MQ_local(recipient)) return false;
        
    // if the connection is authenticated we let the user send the message
    if (MQ_authenticated(connection)) return false;
        
    // a non-local email was submitted, this is not permitted, we send back an alternative response
    // message, because we do not like the response that is normally sent by mailerq
    // we also inform the sender that they should use STARTTLS if they did not do so allready
    if (MQ_secure(connection))
    {
        // the connection was secure, send back an error that the user should first authenticate
        MQ_send(connection, "554-5.7.1 You are not authenticated\r\n", 41);
        MQ_send(connection, "554-5.7.1 It is only possible to submit mails over secure authenticated connections\r\n", 91);
        MQ_send(connection, "554 5.7.1 Please send a valid username and password before you send \"MAIL FROM\"\r\n", 89);
    }
    else
    {
        // the connection was not even secure
        MQ_send(connection, "554-5.7.1 You are not authenticated\r\n", 41);
        MQ_send(connection, "554-5.7.1 It is only possible to submit mails over secure authenticated connections\r\n", 91);
        MQ_send(connection, "554 5.7.1 Please secure the connection with STARTTLS and send a valid username and password before you send \"MAIL FROM\"\r\n", 131);
    }
    
    // return control to the application
    return false;
}

/**
 *  Process an incoming message
 *
 *  @param  ctxt        The context in which the message was received
 *  @param  connection  The connection over which the message came in
 *  @param  message     The actual message that came in
 *  @return do we take responsibility from now
 */
bool mq_smtp_in_message(MQ_Context *ctxt, MQ_Connection *connection, MQ_Message *message)
{    
    // retrieve the user data for this connection
    auto *user = (User *)MQ_connectionData(connection);

    // if no user is given we ignore
    if (user == nullptr) return false;

    // get pointer to message json
    auto *json = MQ_json(message);

    // add the user to the json
    fjson_object_object_add(json, "user", fjson_object_new_string(user->name()));
   
    // we never take responsibility here
    return false;
}

/**
 *  Finalize an incoming connection
 *
 *  @param  context     The context in which the connection was running
 *  @param  connection  The connection that was just closed
 *  @return do we take responsibility here
 */
MQ_EXPORT bool mq_smtp_in_close(MQ_Context *context, MQ_Connection *connection)
{
    // retrieve the user
    auto *user = (User *)MQ_connectionData(connection);

    // clean up memory
    if (user) delete user;

    // we don't take responsibility here
    return false;
}

/**
 *  End of C-style function export
 */
}
