/**
 *  User.h
 *
 *  Class that contains user data  
 *
 *  @copyright 2016 Copernica BV
 */

/**
 *  Include guard
 */
#pragma once

/**
 *  Dependencies
 */
#include <memory>
#include <unistd.h>
#include <crypt.h>
#include <string.h>
#include <shadow.h>

/**
 *  Class that contains userdata
 */
class User
{
private:
    /**
     *  The username
     *  @var  std::string
     */
    std::string _name;

    /**
     *  Entry from the shadow password file
     *  @var struct spwd
     */
    struct spwd *_entry;

    /**
     *  The salt of the password hash
     *  @var  std::string
     */
    char *_salt;
    

public:
    /**
     *  Constructor
     *  @param  name
     *  @param  size
     */
    User(const char *name, size_t size) : 
        _name(name, size), 
        _entry(getspnam(_name.data()))
    {
        // entry must exist
        if (_entry == nullptr) throw std::runtime_error("No such user");
        
        // look up the salt
        _salt = strdup(_entry->sp_pwdp);
        
        // should succeed
        if (_salt == nullptr) throw std::runtime_error("Failed to allocate memory");
        
        // look for the first dollar sign
        auto *dollar = strchr(_salt + 1, '$');
        if (dollar == NULL) throw "Invalid password in file";
        
        // look for the second dollar sign
        dollar = strchr(dollar + 1, '$');
        if (dollar == NULL) throw "Invalid password in file";
        
        // add an end-of-string character to the end of the file
        dollar[1] = 0;
    }
    
    /**
     *  Destructor
     */
    virtual ~User()
    {
        // deallocate salt
        free(_salt);
    }

    /**
     *  Check if the password matches
     *  @param passwd
     *  @return bool
     */
    bool matches(const std::string &passwd) const
    {
        // data we need for encryption
        struct crypt_data data;
        
        // encrypt the user supplied password
        auto *encrypted = crypt_r(passwd.data(), _salt, &data);
        
        // was this a success?
        if (encrypted == nullptr) return false;
        
        // compare the data
        return strcmp(encrypted, _entry->sp_pwdp) == 0;
    }
    
    /**
     *  Expose the user name
     *  @return const char *
     */
    const char *name() const
    {
        return _name.data();
    }
};

